require_relative "spec_helper"
require "sqlite3"
require_relative "../lib/data_broker"

class Parent
  include DataBroker.value_object

  values do
    attribute :id, Integer
  end

  has_many :children
end

class Child
  include DataBroker.value_object

  values do
    attribute :id, Integer
    attribute :name, String
    attribute :parent_id, Integer
  end

  belongs_to :parent
end

class ParentRecord < ActiveRecord::Base
  self.table_name = :parents
  has_many :children, :class_name => "ChildRecord",
    :foreign_key => :parent_id
end

class ChildRecord < ActiveRecord::Base
  self.table_name = :children
  belongs_to :parent, :class_name => "ParentRecord",
    :foreign_key => :parent_id
end

class ParentMapper
  include DataBroker.mapper

  self.mapping = {
    :id => :id
  }

  self.children = {
    :children => :children
  }
end

class ChildMapper
  include DataBroker.mapper

  self.mapping = {
    :id => :id,
    :name => :child_name,
    :parent_id => :parent_id
  }

  self.parents = {
    :parent => :parent
  }
end

describe DataBroker::Mapper do
  let(:mapper) { ChildMapper }
  let(:parent_mapper) { ParentMapper }

  it "infers record if not given" do
    expect(mapper.record).to eq(ChildRecord)
  end

  it "infers model if not given" do
    expect(mapper.model).to eq(Child)
  end

  describe "#all" do
    it "uses the record to fetch all entries" do
      3.times { mapper.record.create }

      objs = mapper.all
      expect(objs.size).to eq(3)
    end

    it "returns objects cast as the model" do
      mapper.record.create
      obj = mapper.all.first

      expect(obj).to be_a(mapper.model)
    end

    it "eager loads if requested" do
      3.times { mapper.record.create }
      expect_any_instance_of(mapper).to receive(:load_parents!).exactly(3).times
      expect_any_instance_of(mapper).to receive(:load_children!).once

      mapper.all(:eager_load => true)
    end
  end

  describe "#find" do
    it "finds the record matching the given primary key" do
      id = mapper.record.create(:child_name => "Team Jacob").id

      obj = mapper.find(id)
      expect(obj.id).to eq(id)
      expect(obj.name).to eq("Team Jacob")
    end

    it "returns the object cast as the model" do
      id = mapper.record.create.id

      obj = mapper.find(id)
      expect(obj).to be_a(mapper.model)
    end

    it "returns nil if the object isn't found" do
      obj = mapper.find(0)
      expect(obj).to be_nil
    end

    it "eager loads has_many relations" do
      parent_id = parent_mapper.record.create.id
      id = mapper.record.create(:parent_id => parent_id)
      obj = mapper.find(id)

      parent = obj.instance_variable_get(:@parent)
      expect(parent).to eq(parent_mapper.find(parent_id))
    end

    it "eager loads belongs_to relations" do
      parent_id = parent_mapper.record.create.id
      id = mapper.record.create(
        :child_name => "Bobby", :parent_id => parent_id
      ).id
      obj = parent_mapper.find(parent_id)

      ids = obj.instance_variable_get(:@child_ids)
      expect(ids).to match_array([id])
    end

    it "doesn't eager load if requested" do
      expect(parent_mapper).to_not receive(:load_children!)
      expect(parent_mapper).to_not receive(:load_parents!)

      parent_id = parent_mapper.record.create.id
      mapper.record.create(
        :child_name => "Bobby", :parent_id => parent_id
      ).id
      parent_mapper.find(parent_id, :eager_load => false)
    end
  end

  describe "#where" do
    it "returns and empty array if conds yield no results" do
      objs = mapper.where(:name => "Team Jacob")
      expect(objs).to be_empty
    end

    it "returns the objects cast as the model" do
      id = mapper.record.create.id
      obj = mapper.where(:id => id).first

      expect(obj).to be_a(mapper.model)
    end

    it "returns the objects matching the conds" do
      id = mapper.record.create(:child_name => "Team Jacob").id
      obj = mapper.where(:id => id).first

      expect(obj.id).to eq(id)
      expect(obj.name).to eq("Team Jacob")
    end

    it "uses the mappings to translate model attrs to db columns" do
      id = mapper.record.create(:child_name => "Team Jacob").id
      obj = mapper.where(:name => "Team Jacob").first

      expect(obj.id).to eq(id)
    end

    it "returns all matches" do
      3.times { |n| mapper.record.create(:child_name => "Team Jacob") }
      objs = mapper.where(:name => "Team Jacob")

      expect(objs.size).to eq(3)
    end

    it "eager loads relations" do
      parent_id = parent_mapper.record.create.id
      id = mapper.record.create(:parent_id => parent_id)
      obj = mapper.where(:id => id).first

      parent = obj.instance_variable_get(:@parent)
      expect(parent).to eq(parent_mapper.find(parent_id))
    end

    it "eager loads belongs_to relations" do
      parent_id = parent_mapper.record.create.id
      id = mapper.record.create(
        :child_name => "Bobby", :parent_id => parent_id
      ).id
      obj = parent_mapper.where(:id => parent_id).first

      ids = obj.instance_variable_get(:@child_ids)
      expect(ids).to match_array([id])
    end

    it "doesn't eager load if requested" do
      expect(parent_mapper).to_not receive(:load_children!)
      expect(parent_mapper).to_not receive(:load_parents!)
      parent_id = parent_mapper.record.create.id
      mapper.record.create(
        :child_name => "Bobby", :parent_id => parent_id
      ).id
      parent_mapper.where(:id => parent_id, :eager_load => false).first
    end
  end

  describe "#save" do
    it "translates a model and saves it as a record" do
      mapper.save(mapper.model.new(:name => "Team Jacob"))
      ds = mapper.record.where(:child_name => "Team Jacob")

      expect(ds).to_not be_empty
    end

    it "returns the model" do
      model = mapper.model.new(:name => "Team Jacob")
      rt = mapper.save(model)

      expect(rt.name).to eq(model.name)
    end

    it "updates an existing model" do
      model = mapper.save(mapper.model.new)
      model = model.with(:name => "Team Jacob")
      rt = mapper.save(model)

      expect(rt.id).to eq(model.id)
    end

    it "doesn't eager load if requested" do
      expect(mapper).to_not receive(:load_children!)
      expect(mapper).to_not receive(:load_parents!)
      mapper.save(mapper.model.new, :eager_load => false)
    end
  end

  describe "#destroy" do
    it "returns nil if the object doesn't exist" do
      obj = mapper.destroy(mapper.model.new(:id => 0))
      expect(obj).to be_nil
    end

    it "returns the destroyed object" do
      model = mapper.save(mapper.model.new)
      rt = mapper.destroy(model)

      expect(model).to eq(rt)
    end

    it "destroys the object" do
      model = mapper.save(mapper.model.new)
      mapper.destroy(model)

      expect(mapper.find(model.id)).to be_nil
    end

    it "eager loads if requests" do
      expect_any_instance_of(mapper).to receive(:load_children!).once
      model = mapper.save(mapper.model.new)
      mapper.destroy(model, :eager_load => true)
    end
  end
end
